/*
 *	$RCSfile: biosio.c $		$Revision: 1.0 $
 *
 *	Bios level interface functions
 *
 * 	Graham E. Kinns		Borland C++ 4.0x
 *
 *	$Log: biosio.c $
 *		Revision 1.0  1995/10/30  21:07:29  Graham
 *		Initial revision
 *
 */
#ifdef RCS
  static const char	RCS_id[] = "$Id: biosio.c 1.0 1995/10/30 21:07:29 Graham Exp $";
#endif

#include <stdio.h>
#include <dos.h>
#include <mem.h>	/*  for movedata  */

#include "global.h"
#include "msdos.h"
#include "biosio.h"

static	BiosErrorDescription	bios_error_desc[] =
{
/*
		00h successful completion
		01h invalid function in AH or invalid parameter
		02h address mark not found
		03h disk write-protected (floppy)
		04h sector not found
		06h disk changed (floppy)
		08h DMA overrun
		09h attempted DMA across 64K boundary
		0Ch unsupported track or invalid media
		10h uncorrectable CRC or ECC error on read
		20h controller failure
		40h seek failed
		80h timeout (not ready)
 */
	{ 0x00,	"No Error" },
	{ 0x01, "Invalid parameter in function call" },
	{ 0x02, "Address mark not found - Blank disk or HD disk in DD drive" },
	{ 0x03, "Disk is write-protected" },
	{ 0x04, "Sector not found - unrecognised format" },
	{ 0x06, "Disk change not caught" },
	{ 0x08, "DMA overrun" },
	{ 0x09, "DMA across 64k page boundary" },
	{ 0x0c, "Unsupported track/invalid media" },
	{ 0x10, "Uncorrected CRC/ECC error - bad sector ?" },
	{ 0x20, "Controller or hardware failure" },
	{ 0x40, "Seek failed - Wrong format ?" },
	{ 0x80, "Timeout - no such drive or no disk in drive" },
	{ 0x00, NULL_CPTR }	/*  Dummy entry to mark end of table  */
};

/*  Disk Bios routines  */

Bool	bios_check_disk_parameters ( const char *fnname, Byte drive, Byte track,
		Byte side, Byte start_sector, Byte num_sectors, MemPtr sector_buf )
{
	if ( drive > BIOS_MAX_DRIVE )
	{
		internal_error( "%s passed drive=%u", fnname, drive );
		return (Bool) TRUE;
	}
	else if ( side > BIOS_MAX_SIDE )
	{
		internal_error( "%s passed side=%u", fnname, side );
		return (Bool) TRUE;
	}
	else if ( track > BIOS_MAX_TRACK )
	{
		internal_error( "%s passed track=%u", fnname, track );
		return (Bool) TRUE;
	}
	else if ( start_sector > BIOS_MAX_SECTOR )
	{
		internal_error( "%s passed sector=%u", fnname, start_sector );
		return (Bool) TRUE;
	}
	else if ( num_sectors == 0x00 )
	{
		internal_error( "%s passed num_sectors=0", fnname );
		return (Bool) TRUE;
	}
	else if ( sector_buf == (MemPtr) NULL )
	{
		internal_error( "%s passed sector_buf = NULL ptr", fnname );
		return (Bool) TRUE;
	}
	else
		return (Bool) FALSE;
}

Byte	bios_reset_drive ( Byte drive )
/*  Attempts to reset passed drive unit using int 0x13 service ah=0x00  */
/*  Returns 0x01 - illegal param - for a bad drive number  */
{
	union REGS	regs;

	DEBUG_PRINTF(( "bios_reset_drive( 0x%02x )", drive ));

	if ( bios_check_disk_parameters( "bios_reset_drive()", drive, 0, 0, 0,
		 1, (MemPtr) "" ))
	{
		return (Byte) BIOS_ERR_BAD_PARAM;
	}

	/*  check its a valid drive unit  */

	regs.h.ah = 0x00;	/*  Int 0x13 Service 0x00, reset drive  */
	regs.h.dl = drive;

	/*  This doesn't seem to return an error regardless  */
	(void) int86( 0x13, &regs, &regs );

	DEBUG_PRINTF(( "\tint 13h ah=00h returned cflag=%d ah=%02x",
			   regs.x.cflag, regs.h.ah ));

	return (Byte) ( regs.x.cflag ? regs.h.ah : 0x00 );
}

Byte	bios_get_last_error ( Byte drive )
{
	union REGS	regs;

	DEBUG_PRINTF(( "bios_get_last_error( 0x%02x )", drive ));

	/*  check its a valid drive unit  */
	if ( bios_check_disk_parameters( "bios_get_last_error()", drive, 0, 0,
		 0, 1, (MemPtr) "" ))
	{
		return (Byte) BIOS_ERR_BAD_PARAM;
	}

	regs.h.ah = 0x01;	/*  Int 13h Service 01h, report last err code  */
	regs.h.dl = drive;

	(void) int86( 0x13, &regs, &regs );

	DEBUG_PRINTF(( "\tint 13h ah=01h returned ah=%02x", regs.h.ah ));

	return (Byte) regs.h.ah;
}

Byte	bios_read_sectors ( Byte drive, Byte track, Byte side,
		Byte start_sector, Byte num_sectors, MemPtr sector_buf )
{
	union REGS	regs;
	struct SREGS	segregs;

	DEBUG_PRINTF(( "bios_read_sectors() Dr %u Si %u Trk %2u StSec %2u NumSecs %2u",
			   drive, side, track, start_sector, num_sectors ));
	DEBUG_PRINTF(( "\tsector_buf = %Fp", sector_buf ));

	if ( bios_check_disk_parameters( "bios_read_sectors()", drive,
		track, side, start_sector, num_sectors, sector_buf ))
	{
		return (Byte) BIOS_ERR_BAD_PARAM;
	}

	regs.h.ah = 0x02;	/*  Int13h Service 02h, read sector  */
	regs.h.al = num_sectors;
	regs.h.dl = drive;
	regs.h.dh = side;
	regs.h.ch = track;
	regs.h.cl = start_sector;
	segregs.es  = FP_SEG( sector_buf );
	regs.x.bx = FP_OFF( sector_buf );

	DEBUG_PRINTF(( "\tsector_buf at es:bx=%04x:%04x", segregs.es, regs.x.bx ));

	(void) int86x( 0x13, &regs, &regs, &segregs );

	DEBUG_PRINTF(( "\tint 13h ah=02h returns cflag=%d ax=%04x", regs.x.cflag, regs.x.ax ));

	return (Byte) regs.h.ah;
}

Byte	bios_write_sectors ( Byte drive, Byte track, Byte side,
		Byte start_sector, Byte num_sectors, MemPtr sector_buf )
{
	union REGS	regs;
	struct SREGS	segregs;

	DEBUG_PRINTF(( "bios_write_sectors() Dr %u Si %u Trk %2u StSec %2u NumSecs %2u",
			   drive, side, track, start_sector, num_sectors ));
	DEBUG_PRINTF(( "\tsector_buf = %Fp", sector_buf ));

	if ( bios_check_disk_parameters( "bios_write_sectors()", drive,
		track, side, start_sector, num_sectors, sector_buf ))
	{
		return (Byte) BIOS_ERR_BAD_PARAM;
	}

	regs.h.ah = 0x03;	/*  Int13h Service 03h, write sector  */
	regs.h.al = num_sectors;
	regs.h.dl = drive;
	regs.h.dh = side;
	regs.h.ch = track;
	regs.h.cl = start_sector;
	segregs.es  = FP_SEG( sector_buf );
	regs.x.bx = FP_OFF( sector_buf );

	DEBUG_PRINTF(( "\tsector_buf at es:bx=%04x:%04x", segregs.es, regs.x.bx ));

	(void) int86x( 0x13, &regs, &regs, &segregs );

	DEBUG_PRINTF(( "\tint 13h ah=03h returns cflag=%d ax=%04x", regs.x.cflag, regs.x.ax ));

	return (Byte) regs.h.ah;
}

Byte	bios_format_track ( Byte drive, Byte track, Byte side, Byte num_sectors,
				MemPtr addr_field_buf )
{
	union REGS	regs;
	struct SREGS	segregs;

	DEBUG_PRINTF(( "bios_format_track() Dr %u Si %u Trk %2u NumSec %2u Buf %Fp",
		       drive, side, track, num_sectors, addr_field_buf ));

	regs.h.ah = 0x05;	/*  Int13h Service 05h, format floppy track  */
	regs.h.al = num_sectors;
	regs.h.dl = drive;
	regs.h.dh = side;
	regs.h.ch = track;
	segregs.es  = FP_SEG( addr_field_buf );
	regs.x.bx = FP_OFF( addr_field_buf );

	DEBUG_PRINTF(( "\taddr_field_buf at es:bx=%04x:%04x", segregs.es, regs.x.bx ));

	(void) int86x( 0x13, &regs, &regs, &segregs );

	DEBUG_PRINTF(( "\tint 13h ah=05h returns cflag=%d ax=%04x", regs.x.cflag, regs.x.ax ));

	return (Byte) regs.h.ah;
}

Byte	bios_get_disk_type ( Byte drive )
{
	union	REGS	regs;

	DEBUG_PRINTF(( "bios_get_disk_type( 0x%02x )", drive ));

	regs.h.ah = 0x15;
	regs.h.dl = drive;

	(void) int86( 0x13, &regs, &regs );

	DEBUG_PRINTF(( "\tint 13h ah=15h returns cflag=%d ax=%04x", regs.x.cflag, regs.x.ax ));

	if ( regs.x.cflag )
	{
		/* An error occured  */
		/* Attempt to reset drive */
		(void) bios_reset_drive( drive );

		return ((Byte) 0xff );
	}

	return ((Byte) regs.h.ah );
}

Bool	bios_hook_dbt ( BiosDBT *dbt )
{
	MemPtr	dbtptr = (MemPtr) (dbt->table);
	Word	current_segment;
	Word	current_offset;
	Word	new_segment;
	Word	new_offset;

	/*  read the present seg:off dbt vector  */
	msdos_get_interrupt( BIOS_DBT_INTERRUPT_VECTOR,
				 &current_segment,
				 &current_offset );

	/*  check we haven't already hooked it  */
	new_segment = DWORD_SEG( FARPTR_TO_DWORD( dbtptr ) );
	new_offset  = DWORD_OFF( FARPTR_TO_DWORD( dbtptr ) );

	DEBUG_PRINTF(( "bios_hook_dbt() Old %04x:%04x New %04x:%04x",
			current_segment, current_offset,
			new_segment, new_offset ));

	if ( new_segment == current_segment &&
		 new_offset  == current_offset  )
		return ( TRUE );

	/*  store the old vector  */
	dbt->orig_segment = current_segment;
	dbt->orig_offset  = current_offset;

	/*  copy the contents of the old table into the new  */
#ifdef __TURBOC__
	movedata( current_segment, current_offset, new_segment, new_offset, BIOS_DBT_SIZE );
	/*          source                           destination  */
#else
	/*  Replacement for movedata  */
	{
		int	i;
		MemPtr	orig, dest;

		orig = MK_FP( current_segment, current_offset );
		dest = MK_FP( new_segment, new_offset );

		for ( i = 0; i < BIOS_DBT_SIZE; i++ )
			/*  this will fail if offset wraps round  */
			*dest++ = *orig++;
	}
#endif

	/*  store the pointer to the new table  */
	msdos_set_interrupt( BIOS_DBT_INTERRUPT_VECTOR, new_segment, new_offset );

	return ( FALSE );
}

Bool	bios_restore_dbt ( BiosDBT *dbt )
/*  restores the Disk base table vector to the original  */
{
	MemPtr	dbtptr = (MemPtr) (dbt->table);
	Word	segment;
	Word	offset;

	/*  read the present interrupt vector  */
	msdos_get_interrupt( BIOS_DBT_INTERRUPT_VECTOR, &segment, &offset );

	DEBUG_PRINTF(( "bios_restore_dbt() Old %04x:%04x New %04x:%04x",
			segment, offset,
			dbt->orig_segment, dbt->orig_offset ));

	/*  check we _have_ already hooked it  */
	if ( segment != DWORD_SEG( FARPTR_TO_DWORD( dbtptr ) ) ||
		 offset  != DWORD_OFF( FARPTR_TO_DWORD( dbtptr ) ) )
		return ( TRUE );

	/*  restore the pointer to the old table  */
	msdos_set_interrupt( BIOS_DBT_INTERRUPT_VECTOR, dbt->orig_segment, dbt->orig_offset );

	return ( FALSE );
}

const char	*bios_error ( Byte drive, Byte error_code )
/*  Returns a string ptr relating to the passed error code  */
{
	static	char			unknown_message_buf[40];
	int						error_table_ptr;
	BiosErrorDescription	*error_ptr;

	if ( error_code == 0x00 )
		error_code = bios_get_last_error( drive );

	/*  Search the table  */

	error_table_ptr = 0;

	while ( (error_ptr = &bios_error_desc[error_table_ptr])->string != NULL_CPTR )
	{
		if ( error_ptr->code == error_code )
		{
			/*  Got a hit  */
			break;
		}

		error_table_ptr++;
	}

	if ( error_ptr->string == NULL_CPTR )
	{
		/*  Couldn't match error code  */

		sprintf( unknown_message_buf, "Unknown bios disk error code (%02x)", error_code );
		return((const char *) unknown_message_buf );
	}

	return ( error_ptr->string );
}
