/*
 *	DDskDemo.c - Demo of Dragon Disk reading functions
 * 
 *	Graham E. Kinns			Oct '95
 *
 *	$Log: ddskdemo.c $
 *		Revision 1.0  1995/10/30  21:08:28  Graham
 *		Initial revision
 *
 */
#ifdef RCS
  static const char	RCS_id[] = "$Id: ddskdemo.c 1.0 1995/10/30 21:08:28 Graham Exp $";
#endif /* RCS */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>	/*  ANSI variable argument function conventions  */
#include <ctype.h>	/*  tolower/upper  */
#include <string.h>

#include "global.h"	/*  Custom project wide definitions  */

#include "biosio.h"

/*  Bug fix for Borland C++ v4.0x to cut out C++ exception handling libs  */
#ifdef __BORLANDC__
#  if __BORLANDC__ >= 0x0450 && __BORLANDC__ <= 0x452
     void _ExceptInit( void ) {}
#  endif
#endif

#define DRAGON_SECTOR_SIZE		256
#define DRAGON_SECTORS_PER_TRACK	18
#define DRAGON_MAX_SECTOR		DRAGON_SECTORS_PER_TRACK
#define DRAGON_DIR_TRACK		20
#define DRAGON_FIRST_BITMAP_SECTOR	1
#define DRAGON_BOOT_TRACK		0
#define DRAGON_BOOT_SECTOR		3
#define DRAGON_BOOT_SIG			"OS"

/*  Function prototypes  */
void	internal_error ( const char *fmt, ... );
Bool	read_dragon_sector ( Byte drive, Byte track, Byte sector, Byte *buf );
Bool	dragon_disk_is_bootable ( Byte drive );
Bool	display_dragon_disk_params ( Byte drive );
int	main ( int argc, char *argv[] );

/*  internal_error may be called from any of library functions  */
void    internal_error ( const char *fmt, ... )
/*  This routine is called whenever an fatal error is encountered
 *  It doesn't return ..
 */
{
        va_list arg_ptr;

        DEBUG_PRINTF(( "internal_error() called" ));

        va_start( arg_ptr, fmt );
        (void)  fprintf( stderr, "Internal error: " );
        (void) vfprintf( stderr, fmt, arg_ptr );
        (void)  fprintf( stderr, "\n" );
        va_end( arg_ptr );

        DEBUG_PRINTF(( "Exiting status 1" ));
        exit( 1 );
}

Bool	read_dragon_sector ( Byte drive, Byte track, Byte sector, Byte *buf )
{
	static Byte	result = 0xff;	/*  defined as static non-0 to reset the drive on very first access  */
	Byte		side = 0;
	int		error_count = 0;
	BiosDBT		dbt;		/*  Struct to hold disk base table info  */

	/*  Calculate the side from the Dragon's method of sector numbering  */
	if ( sector > DRAGON_SECTORS_PER_TRACK )
	{
		sector -= DRAGON_SECTORS_PER_TRACK;
		side = 1;
	}

	/*  NOTE: Should really check here if the far address range of buf
         *  (ie segment:offset) crosses a DMA page boundary i.e. a linear 
         *  address ending in 0x00??0000.  If so, the bios call is documented
	 *  as failing - my bios (MRBios 1.0) doesn't return an error - it
	 *  must adjust internally and returns one sector less in AL ..
	 *  Solution would be to malloc another buffer twice as big as needed
	 *  in which a block can guaranteed to be found which will not cross
	 *  such a boundary.
	 */           

	/*  Now attempt to read the sector  */

	/*  First grab the current disk base table, copy it, change the sector
	 *  size code & replace it's address with the new copy
	 */

	if ( bios_hook_dbt( &dbt ))
		return( TRUE );

        dbt.table[ BIOS_DBT_SECTOR_CODE_OFFSET ] = (Byte) ( DRAGON_SECTOR_SIZE >> 8 );
        dbt.table[ BIOS_DBT_MAX_SECTOR_OFFSET ] = (Byte) ( DRAGON_MAX_SECTOR );

	/*  Call int13h to read the sector - most refs recommend doing this
	 *  3 times with a reset to allow for motor spin up etc.
	 */

	do {
		if ( result )
			bios_reset_drive( drive );

		/*  Read 1 sector from drive,track,side,sector into buf  */
		result = bios_read_sectors( drive, track, side, sector, 1, buf );

		if ( result == 0x06 )
		{
			/*  Disk change 'error'  */

			/*  Ignore it for this program but you need to catch it
			 *  if you want to avoid writing to the wrong disk...
			 */
		}

	} while ( result != 0x00 && ( ++error_count < BIOS_MAX_RETRIES ));

	/*  Put back the pointer to the old Disk Base Table  */
	if ( bios_restore_dbt( &dbt ))
		return( TRUE );

	return ( result ? TRUE : FALSE );
}

Bool	dragon_disk_is_bootable ( Byte drive )
/*  Returns TRUE if the disk in the passed drive number is bootable 
 *  else FALSE.  Returns FALSE on any error.
 */
{
	static const char	boot_sig[] = DRAGON_BOOT_SIG;
	Byte			sector_buf[ DRAGON_SECTOR_SIZE ];

	if ( read_dragon_sector( drive,
	                         DRAGON_BOOT_TRACK,
	                         DRAGON_BOOT_SECTOR,
	                         sector_buf ) )
		return ( FALSE );

	if ( strncmp( boot_sig, (const char *) sector_buf, strlen( boot_sig )) == 0 )
		return ( TRUE );

	return ( FALSE );
}

Bool	display_dragon_disk_params ( Byte drive )
/*  Returns TRUE on error else FALSE  */
{
	Byte	sector_buf[ DRAGON_SECTOR_SIZE ];

	if ( read_dragon_sector( drive,
	                         DRAGON_DIR_TRACK,
	                         DRAGON_FIRST_BITMAP_SECTOR,
	                         sector_buf ) )
		return ( TRUE );

	/*  Check for the valid directory condition from the end of the
	 *  first bitmap sector
	 */

	if ( sector_buf[252] != (Byte) ~sector_buf[254] ||
	     sector_buf[253] != (Byte) ~sector_buf[255] )
	{
		printf( "Not a Dragon Dos disk.\n" );
	}
	else
	{
		printf( "Dragon Dos disk: %02u tracks, %02u sectors.  ", sector_buf[252], sector_buf[253] );
		if ( !dragon_disk_is_bootable( drive ))
			printf( "Not " );
		printf( "Bootable.\n" );
	}

	return ( FALSE );	
}

/*  Program entry point ..  */
int	main ( int argc, char *argv[] )
{
	char	drive_let;	/*  Drive letter A/B/0/1  */
	Byte	drive_num;	/*  Physical unit number 0/1  */

	/*  Opening banner ..  */
	printf( "DDskDemo - Demo of Dragon Dos disk reading - Graham E. Kinns - %s\n\n", __DATE__ );

	/*  Check a drive spec is given on command line  */
	if ( argc != 2 ||
             strlen( argv[1] ) > 2 ||
	     ( argv[1][1] != '\0' && argv[1][1] != ':' )
	   )
	{
		fprintf( stderr, "Usage: %s drive:\n", argv[0] );
		return( 1 );
	}
	
	/*  Process command_line  */
	switch( drive_let = tolower( argv[1][0] ) )
	{
		case '0':	/*  Fall thru  */
		case '1':
			drive_num = drive_let - '0';
			break;

		case 'a':
		case 'b':
			drive_num = drive_let - 'a';
			break;

		default:
			fprintf( stderr, "%s: Unknown drive spec '%s'\n", argv[0], argv[1] );
			return( 1 );
	}

	if ( display_dragon_disk_params( drive_num ))
	{
		/*  Non-zero return - some error occurred
		 *  calling bios_error with a value 0 should determine the 
		 *  last error cause for the given drive
		 */
		fprintf( stderr, "Disk Error: %s\n", bios_error( drive_num, 0 ));
		(void) bios_reset_drive( drive_num );
		return( 1 );
	}

	/*  Reset the drive before leaving ready for next access  */
	(void) bios_reset_drive( drive_num );

	/*  Successful exit  */
	return( 0 );
}

